/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: pfit.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: pfit.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $ */
/* Routines to fit position and flux of the peak in an array  */
#include <math.h>
#include "pfit.h"
#include "interpol.h"

/* internal prototypes */
int momnt (float ara[9][9], float x, float y, int nx, int ny, float momar[6],
   float fblank);
void matvmu (float m[3][3], float vi[3], float vo[3], int n);

int pfit (float a[9][9], float *s, float dx[2], float fblank)
/*--------------------------------------------------------------------*/
/*  Make a parabolic least-squares fit to a 9x9 matrix about the      */
/*  peak value in an array and determine strength and position        */
/*  of maximum.                                                       */
/*   Inputs:                                                          */
/*     a       Data input array[x][y]                                 */
/*     fblank  Value for blanked pixel                                */
/*  Outputs:                                                          */
/*     dx  Position of max relative to a[4][4]                        */
/*     s   Strength of max                                            */
/*     return  0=OK else failed                                       */
/*--------------------------------------------------------------------*/
{
  float  absmax, x, y, temp[6], momar[6], d, st;
  float mat[3][3] = {{0.55555, -0.33333, -0.33333}, {-0.33333, 0.5, 0.0},
		     {-0.33333, 0.0, 0.5}};
  int ix, iy, ixmax, iymax;
  /* default return values */
  dx[0] = 0.0;
  dx[1] = 0.0;
  *s = a[4][4];
  /*  find peak in array    */
  absmax = 0.0; ixmax = -1; iymax = -1;
  for (ix=0; ix<9; ix++) {
    for (iy=0; iy<9; iy++) {
      if ((a[iy][ix]!=fblank) && (fabs(a[iy][ix])>absmax)) 
	{absmax=fabs(a[iy][ix]); ixmax = ix; iymax = iy;}
    }
  }
  /* check for valid data */
  if ((ixmax<0) || (iymax<0)) return 1;
  /*  00, 01, 02, 10, 11, 20 */
  x = ixmax;
  y = iymax;
  /*  next best guess */
  dx[0] = x - 4.0;
  dx[1] = y - 4.0;
  *s = a[iymax][ixmax];

  if (momnt (a, x, y, 3, 3, momar, fblank)) return 1;
  /*  multiply matrix * even moms  yields const & quadratic terms */
  temp[0] = momar[0];
  temp[1] = momar[2];
  temp[2] = momar[5];
  matvmu (mat, temp, &temp[3], 3);
  /*  pick up linear & cross term  */
  temp[0] = momar[1] / 6.;
  temp[1] = momar[3] / 6.;
  temp[2] = momar[4] / 4.;
  /*  offset of peak */
  d = 4.* temp[4] * temp[5] - (temp[2]*temp[2]);
  if (d==0.0) return 2;
  dx[0] = (temp[2]*temp[0] - 2.*temp[1]*temp[4]) / d;
  dx[1] = (temp[2]*temp[1] - 2.*temp[0]*temp[5]) / d;

  /*  interpolate value of peak */
  st = imintp (&a[0][0], 9, 9, 3, dx[0]+4.0, dx[1]+4.0, fblank);
  if (st!=fblank) *s = st;
    
 /* correct wrt center of input array */
  dx[0] = dx[0] + x - 4.0; 
  dx[1] = dx[1] + y - 4.0;

  return 0;
} /* end of pfit */

int momnt (float ara[9][9], float x, float y, int nx, int ny, float momar[6],
	   float fblank)
/*-----------------------------------------------------------------------*/
/*  Calculate all 0th, 1st, and 2nd moments of a nx*ny subarray of ara,  */
/*  centered at x,y.  nx and ny should be odd.                           */
/*  Inputs:                                                              */
/*     ara     Input data array                                          */
/*     x       x-center for moment calculation (0-rel)                   */
/*     y       y-center                                                  */
/*     nx      # of points to include in x-direction. nx                 */
/*             should be odd.  The points will be centered               */
/*             about x (rounded)                                         */
/*     ny      # of points in y-direction                                */
/*  Outputs:                                                             */
/*     momar   00,10,20,01,11,02 yx-moments of ara                       */
/*     return  0 => o.k.                                                 */
/*             1 => subarray doesn't fit in main array                   */
/*-----------------------------------------------------------------------*/
{
  long  ind, i, j, k, nj, indx, indy, iax1, iax2, iay1, iay2;
  float s, t, arg, prod;
  /*      compute loop limits (1-rel)   */
  i = x + 0.5;
  iax1 = i - nx/2;
  iax2 = iax1 + nx - 1;
  i = y + 0.5;
  iay1 = i - ny/2;
  iay2 = iay1+ ny - 1;
  /*      check loop limits             */
  if ((iax1<0) || (iax2>8) || (iay1<0) || (iay2>8)) return 1;
  /*      compute moments               */
  ind = 0;
  for (i = 1; i<=3; i++) {
    nj = 4 - i;
    for (j=1; j<=nj; j++) {
      ind = ind + 1;
      s = 0.0;
      for (indx=iax1; indx<=iax2; indx++) {
	for (indy=iay1; indy<=iay2; indy++) {
	  t = ara[indy][indx];
	  if (t!=fblank) {
	    if (i>1) {
	      prod = 1.0; 
	      arg = indx - x;
	      for (k=1; k<i; k++) prod *= arg;
	      t = t * prod;}     /* ((indx - x)**(i-1)); */
	    if (j>1) {
	      prod = 1.0; 
	      arg = indy - y;
	      for (k=1; k<j; k++) prod *= arg;
	      t = t * prod;}  /* ((indy - y)**(j-1));*/
	    s = s + t;}
	}
      }
      momar[ind-1] = s;
    }
  }
  return 0;
}  /* end of momnt */

void matvmu (float m[3][3], float vi[3], float vo[3], int n)
/*-----------------------------------------------------------------------*/
/*  Matrix-vector multiplication  vo = vi * m                            */
/*  Inputs:                                                              */
/*     m       Input matrix                                              */
/*     vi      Input vector                                              */
/*     n       Array dimension                                           */
/*  Outputs:                                                             */
/*     vo      Output vector                                             */
/*-----------------------------------------------------------------------*/
{
  int  i, j;
  float s;
  
  for (i=0; i<n; i++) {
    s = 0.0;
    for (j=0; j<n; j++) s = s + m[j][i] * vi[j];
    vo[i] = s;
  }
}  /* end of matvmu */




/* */
